import logging
import os
import signal
import socket
import struct
import time
import traceback
from multiprocessing.synchronize import Event
from typing import Callable, Type

from pydantic import BaseModel, ValidationError

from .protocol import Protocol
from .worker import Worker

logger = logging.getLogger(__name__)


CONN_MAX_RETRY = 10
CONN_CHECK_INTERVAL = 1

STAGE_INGRESS = "ingress"
STAGE_EGRESS = "egress"

PROTOCOL_TIMEOUT = 2.0


class Coordinator:
    """
    This private class defines a set of coordination behaviors
    to receive tasks via `Protocol` and process them via `Worker`.
    """

    def __init__(
        self,
        worker: Type[Worker],
        max_batch_size: int,
        stage: str,
        shutdown: Event,
        socket_prefix: str,
        stage_id: int,
        worker_id: int,
        req_schema: Type[BaseModel],
        resp_schema: Type[BaseModel],
    ):
        """Initialize the mosec coordinator

        Args:
            worker (Worker): subclass of `mosec.Worker` implemented by users.
            max_batch_size (int): maximum batch size for this worker.
            stage (str): identifier to distinguish the first and last stages.
            shutdown (Event): `multiprocessing.synchronize.Event` object for shutdown
                IPC.
            socket_prefix (str): prefix for the socket addresses.
            stage_id (int): identification number for worker stages.
            worker_id (int): identification number for worker processes at the same
                stage.
            req_schema (BaseModel): subclass of `pydantic.BaseModel` to define
                input schema for validation, `None` if not required.
            resp_schema (BaseModel): subclass of `pydantic.BaseModel` to define
                output schema for validation, `None` if not required.
        """
        self.worker = worker()
        self.worker._set_mbs(max_batch_size)
        self.worker._set_stage(stage)

        self.req_schema = req_schema
        self.resp_schema = resp_schema

        self.name = f"<{stage_id+1}|{worker.__name__}|{worker_id+1}>"

        self.protocol = Protocol(
            name=self.name,
            addr=os.path.join(socket_prefix, f"ipc_{stage_id}.socket"),
            timeout=PROTOCOL_TIMEOUT,
        )

        # ignore termination & interruption signal
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        self.shutdown = shutdown

        self.init_protocol()
        self.init_worker()
        self.run()

    def exit(self):
        logger.info(f"{self.name} exiting...")
        self.shutdown.set()

    def init_protocol(self):
        """Check socket readiness"""
        retry_count = 0
        while retry_count < CONN_MAX_RETRY and not self.shutdown.is_set():
            if os.path.exists(self.protocol.addr):
                return
            retry_count += 1
            logger.debug(
                f"{self.name} trying to find the socket file: {self.protocol.addr}"
                f" ({retry_count}/{CONN_MAX_RETRY})"
            )
            time.sleep(CONN_CHECK_INTERVAL)
            continue

        logger.error(f"{self.name} cannot find the socket file: {self.protocol.addr}")
        self.exit()

    def init_worker(self):
        """Optional warmup to allocate resources (useful for GPU workload)"""
        if not self.shutdown.is_set():
            pass  # TODO

    def run(self):
        """Maintain the protocol connection and run the coordination"""
        while not self.shutdown.is_set():
            # reconnect if needed
            try:
                self.protocol.open()
            except OSError as err:
                logger.error(f"{self.name} socket connection error: {err}")
                break
            except ConnectionRefusedError as err:
                logger.error(f"{self.name} socket connection refused: {err}")
                break
            except FileNotFoundError:
                logger.error(f"{self.name} socket file not found")
                break

            self.coordinate()

    def get_decoder(self) -> Callable:
        if STAGE_INGRESS in self.worker._stage:
            decoder = self.worker.deserialize

            def validate_decoder(data):
                return self.req_schema.parse_obj(decoder(data))

            return decoder if self.req_schema is None else validate_decoder

        return self.worker._deserialize_ipc

    def get_encoder(self) -> Callable:
        if STAGE_EGRESS in self.worker._stage:
            encoder = self.worker.serialize

            def validate_encoder(data):
                assert isinstance(
                    data, self.resp_schema
                ), f"response {data} is not the instance of {self.resp_schema}"
                return encoder(data)

            return encoder if self.resp_schema is None else validate_encoder

        return self.worker._serialize_ipc

    def coordinate(self):
        """Start coordinating the protocol's communication and worker's forward pass"""
        decoder = self.get_decoder()
        encoder = self.get_encoder()
        while not self.shutdown.is_set():
            try:
                _, ids, payloads = self.protocol.receive()
            except socket.timeout:
                continue
            except (struct.error, OSError) as err:
                logger.error(f"{self.name} socket receive error: {err}")
                break

            try:
                data = [decoder(item) for item in payloads]
                data = (
                    self.worker.forward(data)
                    if self.worker._max_batch_size > 1
                    else (self.worker.forward(data[0]),)
                )
                status = self.protocol.FLAG_OK
                payloads = [encoder(item) for item in data]
                if len(data) != len(payloads):
                    raise ValueError(
                        "returned data doesn't match the input data:"
                        f"input({len(data)})!=output({len(payloads)})"
                    )
            except ValidationError as err:
                err_msg = str(err).replace("\n", " - ")
                logger.info(f"{self.name} validation error: {err_msg}")
                status = self.protocol.FLAG_VALIDATION_ERROR
                payloads = (self.worker.serialize(err.errors()),)
            except Exception:
                logger.warning(traceback.format_exc().replace("\n", " "))
                status = self.protocol.FLAG_INTERNAL_ERROR
                payloads = ("Internal Error".encode(),)

            try:
                self.protocol.send(status, ids, payloads)
            except OSError as err:
                logger.error(f"{self.name} socket send error: {err}")
                break

        self.protocol.close()
        time.sleep(CONN_CHECK_INTERVAL)



import logging

from ._version import __version__  # noqa
from .server import Server
from .worker import Worker

__all__ = ["Server", "Worker"]

# setup library logging
logging.getLogger(__name__).addHandler(logging.NullHandler())



import logging
import multiprocessing as mp
import signal
import subprocess
from multiprocessing.synchronize import Event
from pathlib import Path
from time import monotonic, sleep
from typing import List, Optional, Tuple, Type, Union

import pkg_resources
from pydantic import BaseModel, BaseSettings, conint, validate_arguments

from .coordinator import STAGE_EGRESS, STAGE_INGRESS, Coordinator
from .utils import Settings
from .worker import Worker

logger = logging.getLogger(__name__)

AtLeastOne: Type[int] = conint(strict=True, ge=1)

GUARD_CHECK_INTERVAL = 1
NEW_PROCESS_METHOD = {"spawn", "fork"}


class Server:
    def __init__(
        self, req_schema: Type[BaseModel] = None, resp_schema: Type[BaseModel] = None
    ):
        self._req_schema = req_schema
        self._resp_schema = resp_schema

        self._worker_cls: List[Type[Worker]] = []
        self._worker_num: List[AtLeastOne] = []  # type: ignore
        self._worker_mbs: List[AtLeastOne] = []  # type: ignore

        self._coordinator_ctx: List[str] = []
        self._coordinator_pools: List[List[Union[mp.Process, None]]] = []
        self._coordinator_shutdown: List[List[Union[Event, None]]] = []

        self._controller_process: Optional[mp.Process] = None

        self._configs: BaseSettings = Settings()

        self._server_shutdown: bool = False
        signal.signal(signal.SIGTERM, self._terminate)
        signal.signal(signal.SIGINT, self._terminate)

    def _validate(self):
        assert len(self._worker_cls) > 0, (
            "no worker registered\n"
            "help: use `.append_worker(...)` to register at least one worker"
        )

    def _parse_args(self):
        # directly update self._configs
        # serialize configs into env for controller to read
        pass  # TODO

    def _start_controller(self):
        """Subprocess to start controller program"""
        if not self._server_shutdown:
            path = Path(pkg_resources.resource_filename("mosec", "bin"), "mosec")
            self._controller_process = subprocess.Popen([path])

    def _terminate(self, signum, framestack):
        logger.info(f"[{signum}] terminating server [{framestack}] ...")
        self._server_shutdown = True

    @staticmethod
    def _clean_pools(
        processes: List[Union[mp.Process, None]],
        events: List[Union[Event, None]],
    ) -> Tuple[List[Union[mp.Process, None]], List[Union[Event, None]]]:
        for i, (p, e) in enumerate(zip(processes, events)):
            if p is None or p.exitcode is not None:
                processes[i] = None
                events[i] = None
        return processes, events

    def _manage_coordinators(self):
        first = True
        while not self._server_shutdown:
            for stage_id, (w_cls, w_num, w_mbs, c_ctx) in enumerate(
                zip(
                    self._worker_cls,
                    self._worker_num,
                    self._worker_mbs,
                    self._coordinator_ctx,
                )
            ):
                # for every sequential stage
                (
                    self._coordinator_pools[stage_id],
                    self._coordinator_shutdown[stage_id],
                ) = self._clean_pools(
                    self._coordinator_pools[stage_id],
                    self._coordinator_shutdown[stage_id],
                )

                if all(self._coordinator_pools[stage_id]):
                    # this stage is healthy
                    continue

                if not first and not any(self._coordinator_pools[stage_id]):
                    # this stage might contain bugs
                    self._terminate(
                        1,
                        f"all workers at stage {stage_id+1} exited;"
                        " please check for bugs or socket connection issues",
                    )
                    break

                stage = ""
                if stage_id == 0:
                    stage += STAGE_INGRESS
                    req_schema = self._req_schema
                else:
                    req_schema = None
                if stage_id == len(self._worker_cls) - 1:
                    stage += STAGE_EGRESS
                    resp_schema = self._resp_schema
                else:
                    resp_schema = None

                for worker_id in range(w_num):
                    # for every worker in each stage
                    if self._coordinator_pools[stage_id][worker_id] is not None:
                        continue
                    shutdown = mp.get_context(c_ctx).Event()
                    coordinator_process = mp.get_context(c_ctx).Process(
                        target=Coordinator,
                        args=(
                            w_cls,
                            w_mbs,
                            stage,
                            shutdown,
                            self._configs.socket_prefix,
                            stage_id,
                            worker_id,
                            req_schema,
                            resp_schema,
                        ),
                        daemon=True,
                    )
                    coordinator_process.start()
                    self._coordinator_pools[stage_id][worker_id] = coordinator_process
                    self._coordinator_shutdown[stage_id][worker_id] = shutdown
            first = False
            if self._controller_process:
                ctr_exitcode = self._controller_process.poll()
                if ctr_exitcode:
                    self._terminate(
                        ctr_exitcode,
                        f"mosec controller exited on error: {ctr_exitcode}",
                    )
            sleep(GUARD_CHECK_INTERVAL)

    def _halt(self):
        """Graceful shutdown"""
        # terminate controller first and wait for a graceful period
        if self._controller_process:
            self._controller_process.terminate()
            graceful_period = monotonic() + self._configs.timeout / 1000
            while monotonic() < graceful_period:
                ctr_exitcode = self._controller_process.poll()
                if ctr_exitcode:
                    self._terminate(
                        ctr_exitcode,
                        f"mosec controller halted on error: {ctr_exitcode}",
                    )
                    break
                sleep(0.1)

        for shutdown_events in self._coordinator_shutdown:
            for event in shutdown_events:
                if event:
                    event.set()
        logger.info("mosec server exited. see you.")

    @validate_arguments
    def append_worker(
        self,
        worker: Type[Worker],
        num: AtLeastOne = 1,  # type: ignore
        max_batch_size: AtLeastOne = 1,  # type: ignore
        start_method: str = "spawn",
    ):
        """Sequentially add workers to be invoked in a pipelined manner"""
        assert (
            start_method in NEW_PROCESS_METHOD
        ), f"start method needs to be one of {NEW_PROCESS_METHOD}"

        self._worker_cls.append(worker)
        self._worker_num.append(num)
        self._worker_mbs.append(max_batch_size)
        self._coordinator_ctx.append(start_method)
        self._coordinator_pools.append([None] * num)
        self._coordinator_shutdown.append([None] * num)

    def run(self):
        """Run the mosec model server!"""
        self._validate()
        self._parse_args()
        self._start_controller()
        self._manage_coordinators()
        self._halt()



import json
import pickle

from pydantic.json import pydantic_encoder


class Worker:
    """
    This public class defines the mosec worker interface. It provides
    default IPC de/serialize methods, stores the worker stage and
    maximum batch size, and implements the forward pass computation.
    """

    def __init__(self):
        self._stage = None
        self._max_batch_size = 1

    def _serialize_ipc(self, data):
        """Define IPC serialize method"""
        return pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL)

    def _deserialize_ipc(self, data):
        """Define IPC deserialize method"""
        return pickle.loads(data)

    def _set_stage(self, stage):
        self._stage = stage

    def _set_mbs(self, mbs):
        self._max_batch_size = mbs

    def serialize(self, data):
        """Define service egress pack method, overridable"""
        return json.dumps(data, indent=2, default=pydantic_encoder).encode()

    def deserialize(self, data):
        """Define service ingress unpack method, overridable"""
        return json.loads(data) if data else {}

    def forward(self, data):
        """Define worker's computation, must be overridden by all subclasses"""
        raise NotImplementedError



import logging
import socket
import struct
from itertools import zip_longest

logger = logging.getLogger(__name__)


class Protocol:
    """
    This private class implements the client-side
    protocol based on uds to communicate with the
    server hosted on mosec controller side.
    """

    # byte formats (https://docs.python.org/3/library/struct.html#format-characters)
    FORMAT_FLAG = "!H"
    FORMAT_BATCH = "!H"
    FORMAT_ID = "!I"
    FORMAT_LENGTH = "!I"

    # flags
    FLAG_OK = 1  # 200
    FLAG_BAD_REQUEST = 2  # 400
    FLAG_VALIDATION_ERROR = 4  # 422
    FLAG_INTERNAL_ERROR = 8  # 500

    # lengths
    LENGTH_TASK_FLAG = 2
    LENGTH_TASK_BATCH = 2
    LENGTH_TASK_ID = 4
    LENGTH_TASK_BODY_LEN = 4

    def __init__(
        self,
        name: str,
        addr: str,
        timeout: float = 2.0,
    ):
        """Initialize the protocol client

        Args:
            name (str): name of its belonging coordinator.
            addr (str): unix domain socket address in file system's namespace.
            timeout (float, optional): socket timeout. Defaults to 2.0 seconds.
        """
        self.socket = socket.socket(
            socket.AF_UNIX,
            socket.SOCK_STREAM,
        )
        self.socket.settimeout(timeout)
        self.name = name
        self.addr = addr

    def receive(self):
        """Receive tasks from the server"""
        flag = self.socket.recv(self.LENGTH_TASK_FLAG)
        batch_size_bytes = self.socket.recv(self.LENGTH_TASK_BATCH)
        batch_size = struct.unpack(self.FORMAT_BATCH, batch_size_bytes)[0]
        ids, payloads = [], []

        while batch_size > 0:
            batch_size -= 1
            id_bytes = self.socket.recv(self.LENGTH_TASK_ID)
            length_bytes = self.socket.recv(self.LENGTH_TASK_BODY_LEN)
            length = struct.unpack(self.FORMAT_LENGTH, length_bytes)[0]
            payload = _recv_all(self.socket, length)
            ids.append(id_bytes)
            payloads.append(payload)

        logger.debug(f"{self.name} received {len(ids)} tasks with ids: {ids}")
        return flag, ids, payloads

    def send(self, flag, ids, payloads):
        """Send results to the server"""
        data = bytearray()
        data.extend(struct.pack(self.FORMAT_FLAG, flag))
        batch_size = len(ids)
        data.extend(struct.pack(self.FORMAT_BATCH, batch_size))
        if batch_size > 0:
            for task_id, payload in zip_longest(ids, payloads, fillvalue=payloads[0]):
                length = struct.pack(self.FORMAT_LENGTH, len(payload))
                data.extend(task_id)
                data.extend(length)
                data.extend(payload)

        self.socket.sendall(data)
        logger.debug(f"{self.name} sent {len(ids)} tasks with ids: {ids}")

    def open(self):
        """Open the socket connection"""
        self.socket.connect(self.addr)
        logger.info(f"{self.name} socket connected to {self.addr}")

    def close(self):
        """Close the socket connection"""
        self.socket.close()
        logger.info(f"{self.name} socket closed")


def _recv_all(conn, length):
    buffer = bytearray(length)
    mv = memoryview(buffer)
    size = 0
    while size < length:
        packet = conn.recv_into(mv)
        mv = mv[packet:]
        size += packet
    return buffer



import os
import shutil
import subprocess
from io import open
from os import path, sep

from setuptools import Extension, find_packages, setup  # type: ignore
from setuptools.command.build_ext import build_ext as _build_ext  # type: ignore

here = path.abspath(path.dirname(__file__))
PACKAGE_NAME = "mosec"

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()

with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    requires = [req.strip() for req in f if req]


def get_version():
    """Use rust package version as the single source for versioning"""
    version = (
        subprocess.check_output(["cargo", "pkgid"]).decode().strip().split("#")[-1]
    )

    version_str = '__version__ = "{}"'.format(version)

    # update py version
    with open("mosec/_version.py", "w") as f:
        f.write(f"{version_str}\n")
    return version


class RustExtension(Extension):
    """Custom Extension class for rust"""


ext_modules = []

if os.getenv("PRODUCTION_MODE"):
    ext_modules.append(RustExtension(name="mosec.bin", sources=["src/*"]))


class RustBuildExt(_build_ext):
    """Custom build extension class for rust"""

    def run(self):
        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext: Extension):
        if not isinstance(ext, RustExtension):
            return super().build_extension(ext)

        libpath = ext.name.replace(".", sep)
        build_libpath = path.join(self.build_lib, libpath)
        os.makedirs(
            os.path.dirname(path.join(build_libpath, PACKAGE_NAME)), exist_ok=True
        )
        rust_target = os.getenv("RUST_TARGET")
        build_cmd = ["cargo", "build", "--release"]
        if rust_target is not None:
            build_cmd += ["--target", rust_target]

        print(f"running rust cargo package build: {build_cmd}")
        errno = subprocess.call(build_cmd)

        assert errno == 0, "Error occurred while building rust binary"

        # package the binary
        if rust_target is not None:
            target_dir = path.join("target", rust_target, "release", PACKAGE_NAME)
        else:
            target_dir = path.join("target", "release", PACKAGE_NAME)
        shutil.copy(target_dir, path.join(build_libpath, PACKAGE_NAME))

        if self.inplace:
            os.makedirs(os.path.dirname(libpath), exist_ok=True)
            shutil.copy(build_libpath, libpath)


setup(
    name=PACKAGE_NAME,
    version=get_version(),
    author="Keming Yang",
    author_email="kemingy94@gmail.com",
    description="Model Serving made Efficient in the Cloud.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/mosecorg/mosec",
    packages=find_packages(exclude=["examples*", "tests*"]),
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.6",
    install_requires=requires,
    extras_require={
        "dev": [
            "pytest>=6",
            "pytest-mock>=3.5",
            "mypy>=0.910",
            "flake8>=3.8",
            "black>=20.8b1",
            "isort>=5.6",
            "autoflake>=1.4",
            "msgpack>=1.0.2",
            "pre-commit>=2.15.0",
        ],
    },
    zip_safe=False,
    entry_points={
        "console_scripts": [],
    },
    ext_modules=ext_modules,
    cmdclass=dict(build_ext=RustBuildExt),
)
